"""Test image converter with identical basenames"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp


@pytest.mark.sphinx('latex', testroot='ext-imgmockconverter')
def test_ext_imgmockconverter(app: SphinxTestApp) -> None:
    app.build(force_all=True)

    content = (app.outdir / 'projectnamenotset.tex').read_text(encoding='utf8')

    # check identical basenames give distinct files
    assert '\\sphinxincludegraphics{{svgimg}.pdf}' in content
    assert '\\sphinxincludegraphics{{svgimg1}.pdf}' in content
    assert not (app.outdir / 'svgimg.svg').exists()
    assert (app.outdir / 'svgimg.pdf').exists()
    assert (app.outdir / 'svgimg1.pdf').exists()



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock

# NEVER import those objects from sphinx.ext.autodoc directly
from sphinx.ext.autodoc.directive import DocumenterBridge, process_documenter_options
from sphinx.util.docutils import LoggingReporter

if TYPE_CHECKING:
    from typing import Any

    from docutils.statemachine import StringList

    from sphinx.application import Sphinx


def do_autodoc(
    app: Sphinx,
    objtype: str,
    name: str,
    options: dict[str, Any] | None = None,
) -> StringList:
    options = {} if options is None else options.copy()
    if not app.env.current_document.docname:
        app.env.current_document.docname = 'index'  # set dummy docname
    doccls = app.registry.documenters[objtype]
    docoptions = process_documenter_options(doccls, app.config, options)
    state = Mock()
    state.document.settings.tab_width = 8
    bridge = DocumenterBridge(app.env, LoggingReporter(''), docoptions, 1, state)
    documenter = doccls(bridge, name)
    documenter.generate()
    return bridge.result



"""Tests sphinx.util.rst functions."""

from __future__ import annotations

from docutils.statemachine import StringList
from jinja2 import Environment

from sphinx.util.rst import append_epilog, escape, heading, prepend_prolog, textwidth


def test_escape() -> None:
    assert escape(':ref:`id`') == r'\:ref\:\`id\`'
    assert escape('footnote [#]_') == r'footnote \[\#\]\_'
    assert escape('sphinx.application') == r'sphinx.application'
    assert escape('.. toctree::') == r'\.. toctree\:\:'


def test_append_epilog() -> None:
    epilog = 'this is rst_epilog\ngood-bye reST!'
    content = StringList(
        ['hello Sphinx world', 'Sphinx is a document generator'],
        'dummy.rst',
    )
    append_epilog(content, epilog)

    assert list(content.xitems()) == [
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
        ('dummy.rst', 2, ''),
        ('<rst_epilog>', 0, 'this is rst_epilog'),
        ('<rst_epilog>', 1, 'good-bye reST!'),
    ]


def test_prepend_prolog() -> None:
    prolog = 'this is rst_prolog\nhello reST!'
    content = StringList(
        [
            ':title: test of SphinxFileInput',
            ':author: Sphinx team',
            '',
            'hello Sphinx world',
            'Sphinx is a document generator',
        ],
        'dummy.rst',
    )
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('dummy.rst', 0, ':title: test of SphinxFileInput'),
        ('dummy.rst', 1, ':author: Sphinx team'),
        ('<generated>', 0, ''),
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 2, ''),
        ('dummy.rst', 3, 'hello Sphinx world'),
        ('dummy.rst', 4, 'Sphinx is a document generator'),
    ]


def test_prepend_prolog_with_CR() -> None:
    # prolog having CR at tail
    prolog = 'this is rst_prolog\nhello reST!\n'
    content = StringList(
        ['hello Sphinx world', 'Sphinx is a document generator'],
        'dummy.rst',
    )
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
    ]


def test_prepend_prolog_without_CR() -> None:
    # prolog not having CR at tail
    prolog = 'this is rst_prolog\nhello reST!'
    content = StringList(
        ['hello Sphinx world', 'Sphinx is a document generator'],
        'dummy.rst',
    )
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 0, 'hello Sphinx world'),
        ('dummy.rst', 1, 'Sphinx is a document generator'),
    ]


def test_prepend_prolog_with_roles_in_sections() -> None:
    prolog = 'this is rst_prolog\nhello reST!'
    content = StringList(
        [
            ':title: test of SphinxFileInput',
            ':author: Sphinx team',
            '',  # this newline is required
            ':mod:`foo`',
            '----------',
            '',
            'hello',
        ],
        'dummy.rst',
    )
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('dummy.rst', 0, ':title: test of SphinxFileInput'),
        ('dummy.rst', 1, ':author: Sphinx team'),
        ('<generated>', 0, ''),
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 2, ''),
        ('dummy.rst', 3, ':mod:`foo`'),
        ('dummy.rst', 4, '----------'),
        ('dummy.rst', 5, ''),
        ('dummy.rst', 6, 'hello'),
    ]


def test_prepend_prolog_with_roles_in_sections_with_newline() -> None:
    # prologue with trailing line break
    prolog = 'this is rst_prolog\nhello reST!\n'
    content = StringList([':mod:`foo`', '-' * 10, '', 'hello'], 'dummy.rst')
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 0, ':mod:`foo`'),
        ('dummy.rst', 1, '----------'),
        ('dummy.rst', 2, ''),
        ('dummy.rst', 3, 'hello'),
    ]


def test_prepend_prolog_with_roles_in_sections_without_newline() -> None:
    # prologue with no trailing line break
    prolog = 'this is rst_prolog\nhello reST!'
    content = StringList([':mod:`foo`', '-' * 10, '', 'hello'], 'dummy.rst')
    prepend_prolog(content, prolog)

    assert list(content.xitems()) == [
        ('<rst_prolog>', 0, 'this is rst_prolog'),
        ('<rst_prolog>', 1, 'hello reST!'),
        ('<generated>', 0, ''),
        ('dummy.rst', 0, ':mod:`foo`'),
        ('dummy.rst', 1, '----------'),
        ('dummy.rst', 2, ''),
        ('dummy.rst', 3, 'hello'),
    ]


def test_textwidth() -> None:
    assert textwidth('Hello') == 5
    assert textwidth('русский язык') == 12
    assert textwidth('русский язык', 'WFA') == 23  # Cyrillic are ambiguous chars


def test_heading() -> None:
    env = Environment(autoescape=True)
    env.extend(language=None)

    assert heading(env, 'Hello') == 'Hello\n====='
    assert heading(env, 'Hello', 1) == 'Hello\n====='
    assert heading(env, 'Hello', 2) == 'Hello\n-----'
    assert heading(env, 'Hello', 3) == 'Hello\n~~~~~'
    assert heading(env, 'русский язык', 1) == 'русский язык\n============'

    # language=ja: ambiguous
    env.language = 'ja'  # type: ignore[attr-defined]
    assert heading(env, 'русский язык', 1) == 'русский язык\n======================='



"""Test the Pygments highlighting bridge."""

from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pygments
import pytest
from pygments.formatters.html import HtmlFormatter
from pygments.lexer import RegexLexer
from pygments.token import Name, Text

from sphinx.highlighting import PygmentsBridge

if TYPE_CHECKING:
    from sphinx.testing.util import SphinxTestApp

if tuple(map(int, pygments.__version__.split('.')[:2])) < (2, 18):
    from pygments.formatter import Formatter

    Formatter.__class_getitem__ = classmethod(lambda cls, name: cls)  # type: ignore[attr-defined]


class MyLexer(RegexLexer):
    name = 'testlexer'

    tokens = {
        'root': [
            ('a', Name),
            ('b', Text),
        ],
    }


class MyFormatter(HtmlFormatter[str]):
    def format(self, tokensource, outfile):
        for tok in tokensource:
            outfile.write(tok[1])


class ComplainOnUnhighlighted(PygmentsBridge):
    def unhighlighted(self, source):
        raise AssertionError('should highlight %r' % source)


@pytest.mark.sphinx('html', testroot='root')
def test_add_lexer(app: SphinxTestApp) -> None:
    app.add_lexer('test', MyLexer)

    bridge = PygmentsBridge('html')
    ret = bridge.highlight_block('ab', 'test')
    assert '<span class="n">a</span>b' in ret


def test_detect_interactive() -> None:
    bridge = ComplainOnUnhighlighted('html')
    blocks = [
        """
        >>> testing()
        True
        """,
    ]
    for block in blocks:
        ret = bridge.highlight_block(block.lstrip(), 'python')
        assert ret.startswith('<div class="highlight">')


def test_lexer_options() -> None:
    bridge = PygmentsBridge('html')
    ret = bridge.highlight_block('//comment', 'php', opts={'startinline': True})
    assert '<span class="c1">//comment</span>' in ret


def test_set_formatter() -> None:
    PygmentsBridge.html_formatter = MyFormatter
    try:
        bridge = PygmentsBridge('html')
        ret = bridge.highlight_block('foo\n', 'python')
        assert ret == 'foo\n'
    finally:
        PygmentsBridge.html_formatter = HtmlFormatter


@mock.patch('sphinx.highlighting.logger')
def test_default_highlight(logger):
    bridge = PygmentsBridge('html')

    # default: highlights as python3
    ret = bridge.highlight_block('print "Hello sphinx world"', 'default')
    assert ret == (
        '<div class="highlight"><pre><span></span><span class="nb">print</span> '
        '<span class="s2">&quot;Hello sphinx world&quot;</span>\n</pre></div>\n'
    )

    # default: fallbacks to none if highlighting failed
    ret = bridge.highlight_block('reST ``like`` text', 'default')
    assert ret == (
        '<div class="highlight"><pre><span></span>reST ``like`` text\n</pre></div>\n'
    )

    # python: highlights as python3
    ret = bridge.highlight_block('print("Hello sphinx world")', 'python')
    assert ret == (
        '<div class="highlight"><pre><span></span><span class="nb">print</span>'
        '<span class="p">(</span>'
        '<span class="s2">&quot;Hello sphinx world&quot;</span>'
        '<span class="p">)</span>\n</pre></div>\n'
    )

    # python3: highlights as python3
    ret = bridge.highlight_block('print("Hello sphinx world")', 'python3')
    assert ret == (
        '<div class="highlight"><pre><span></span><span class="nb">print</span>'
        '<span class="p">(</span>'
        '<span class="s2">&quot;Hello sphinx world&quot;</span>'
        '<span class="p">)</span>\n</pre></div>\n'
    )

    # python: raises error if highlighting failed
    ret = bridge.highlight_block('reST ``like`` text', 'python')
    logger.warning.assert_called_with(
        'Lexing literal_block %r as "%s" resulted in an error at token: %r. '
        'Retrying in relaxed mode.',
        'reST ``like`` text',
        'python',
        '`',
        type='misc',
        subtype='highlighting_failure',
        location=None,
    )
